#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

# All Certificate Information (root and intermediate) in CCADB (CSV)
# utility functions.
# https://www.ccadb.org/resources
# https://ccadb.my.salesforce-sites.com/ccadb/AllCertificateRecordsCSVFormatv2

from datetime import datetime, timezone
from enum import Flag, auto
from typing import Tuple, Union

_FMT = '%Y.%m.%d %z'

__all__ = [
    'valid_from', 'valid_to', 'valid_from_to', 'revoked',
    'TrustBits', 'derived_trust_bits_list', 'derived_trust_bits_flag',
    'derived_trust_bits',
]


class TrustBits(Flag):
    NONE = 0
    OTHER = auto()
    CLIENT_AUTHENTICATION = auto()
    CODE_SIGNING = auto()
    DOCUMENT_SIGNING = auto()
    OCSP_SIGNING = auto()
    SECURE_EMAIL = auto()
    SERVER_AUTHENTICATION = auto()
    TIME_STAMPING = auto()


# https://www.ccadb.org/cas/fields#formula-fields
TrustBitsMap = {
    'Client Authentication': TrustBits.CLIENT_AUTHENTICATION,
    'Code Signing': TrustBits.CODE_SIGNING,
    'Document Signing': TrustBits.DOCUMENT_SIGNING,
    'OCSP Signing': TrustBits.OCSP_SIGNING,
    'Secure Email': TrustBits.SECURE_EMAIL,
    'Server Authentication': TrustBits.SERVER_AUTHENTICATION,
    'Time Stamping': TrustBits.TIME_STAMPING,
}

TrustBitsMap2 = {v.name: v for v in TrustBitsMap.values()}


def _now():
    now = datetime.now(timezone.utc)

    return now


def valid_from(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Valid From (GMT)'
    valid_from = datetime.strptime(row[k] + ' +0000', _FMT)
    if _now() < valid_from:
        x = 'Not yet valid (valid from %s)' % row[k]
        return False, x

    return True, None


def valid_to(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Valid To (GMT)'
    valid_to = datetime.strptime(row[k] + ' +0000', _FMT)
    if valid_to < _now():
        x = 'Expired (valid to %s)' % row[k]
        return False, x

    return True, None


def valid_from_to(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    return valid_from(row) and valid_to(row)


def revoked(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Revocation Status'
    if row[k] not in ['', 'Not Revoked']:
        return True, row[k]

    return False, None


def derived_trust_bits_list(row: dict[str, str]) -> list[str]:
    k = 'Derived Trust Bits'
    derived_trust_bits = row[k]
    if not derived_trust_bits:
        return []

    return derived_trust_bits.split(';')


def derived_trust_bits_flag(values: list[str]) -> TrustBits:
    bits = TrustBits.NONE
    for x in values:
        if x in TrustBitsMap:
            bits = bits | TrustBitsMap[x]
        else:
            bits = bits | TrustBits.OTHER

    return bits


def derived_trust_bits(row: dict[str, str]) -> TrustBits:
    x = derived_trust_bits_list(row)
    return derived_trust_bits_flag(x)



#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

# crt.sh API interface

import aiohttp
import asyncio

URL = 'https://crt.sh'


class ArgsError(Exception):
    pass


class CrtShApi:
    def __init__(self, *,
                 timeout=None,
                 headers=None):
        self.url = URL

        if isinstance(timeout, tuple):
            if len(timeout) != 2:
                raise ArgsError('timeout tuple length must be 2')
            x = aiohttp.ClientTimeout(sock_connect=timeout[0],
                                      sock_read=timeout[1])
        else:
            x = aiohttp.ClientTimeout(total=timeout)

        self.session = aiohttp.ClientSession(timeout=x,
                                             headers=headers)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.session.close()

    async def content(self, *, resp=None):
        if resp is None:
            raise ArgsError('missing resp')

        if resp.status != 200:
            return None, None

        filename = None
        if (resp.content_disposition is not None and
           resp.content_disposition.filename is not None):
            filename = resp.content_disposition.filename

        x = await resp.text()

        return filename, x

    async def download(self, *,
                       id=None,
                       query_string=None):
        path = '/'
        url = self.url + path

        data = {}
        if id is not None:
            data['d'] = id
        if query_string is not None:
            data.update(query_string)

        kwargs = {
            'url': url,
            'data': data,
        }

        resp = await self.session.post(**kwargs)

        return resp



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
import io
import logging
import os
import sys
import tarfile
import time

try:
    import pan.xapi
except ImportError:
    print('Install pan-python: https://pypi.org/project/pan-python/',
          file=sys.stderr)
    sys.exit(1)

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import (title, __version__)

args = None


def main():
    global args
    args = parse_args()

    logger = logging.getLogger(pan.xapi.__name__)
    if args.xdebug == 3:
        logger.setLevel(pan.xapi.DEBUG3)
    elif args.xdebug == 2:
        logger.setLevel(pan.xapi.DEBUG2)
    elif args.xdebug == 1:
        logger.setLevel(pan.xapi.DEBUG1)

    log_format = '%(name)s: %(message)s'
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    asyncio.run(main_loop())

    sys.exit(0)


async def main_loop():
    try:
        xapi = pan.xapi.PanXapi(tag=args.tag,
                                debug=args.xdebug)
    except pan.xapi.PanXapiError as e:
        print('pan.xapi.PanXapi:', e, file=sys.stderr)
        sys.exit(1)

    create_archive(test=True)

    store = {}
    for cert_name, filename in get_trusted_certs(xapi):
        if args.verbose:
            print('Exporting %s => %s' % (cert_name, filename))
        _, content = export_cert(xapi, cert_name)
        if content is not None:
            store[filename] = content

    total = create_archive(store)
    print('Exported %d PAN-OS trusted CAs to %s' % (total, args.certs))


def get_trusted_certs(xapi):
    kwargs = {'xpath': '/config/predefined/trusted-root-ca/entry'}
    api_request(xapi, xapi.get, kwargs, 'success', '19')
    entries = xapi.element_root.findall('./result/entry')

    for entry in entries:
        name = entry.attrib['name']
        x = entry.find('./filename')
        filename = x.text
        yield name, filename


def export_cert(xapi, name):
    kwargs = {
        'category': 'certificate',
        'extra_qs': {
            'certificate-name': name,
            'format': 'pem',
            'include-key': 'no',
        },
    }
    api_request(xapi, xapi.export, kwargs, 'success')
    if xapi.export_result is None:
        print("Can't export %s" % name, file=sys.stderr)
        return None, None

    return xapi.export_result['file'], xapi.export_result['content']


def api_request(xapi, func, kwargs, status=None, status_code=None):
    try:
        func(**kwargs)
    except pan.xapi.PanXapiError as e:
        print('%s: %s: %s' % (func.__name__, kwargs, e),
              file=sys.stderr)
        sys.exit(1)

    if status is not None and not s1_in_s2(xapi.status, status):
        print('%s: %s: status %s != %s' %
              (func.__name__, kwargs,
               xapi.status, status),
              file=sys.stderr)
        sys.exit(1)

    if (status_code is not None and
       not s1_in_s2(xapi.status_code, status_code)):
        print('%s: %s: status_code %s != %s' %
              (func.__name__, kwargs,
               xapi.status_code, status_code),
              file=sys.stderr)
        sys.exit(1)


def s1_in_s2(s1, s2):
    if isinstance(s2, str):
        return s1 == s2
    elif isinstance(s2, list):
        return s1 in s2
    else:
        raise ValueError('Invalid type for s2. '
                         'Must be a string or a list of strings.')


def create_archive(store=None, test=False):
    try:
        with tarfile.open(name=args.certs, mode='w:gz') as tar:
            if test:
                return
            total = 0
            for filename in sorted(store.keys()):
                member = tarfile.TarInfo(name=filename)
                member.size = len(store[filename])
                member.mtime = time.time()
                f = io.BytesIO(store[filename])
                tar.addfile(member, fileobj=f)
                total += 1
    except (tarfile.TarError, OSError) as e:
        print("tarfile %s: %s" % (args.certs, e), file=sys.stderr)
        sys.exit(1)

    return total


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='export PAN-OS trusted CAs')
    parser.add_argument('--tag', '-t',
                        required=True,
                        help='.panrc tagname')
    x = 'root-store.tgz'
    parser.add_argument('--certs',
                        default=x,
                        metavar='PATH',
                        help='PAN-OS trusted CAs archive path'
                        ' (default: %s)' % x)
    parser.add_argument('--xdebug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='pan.xapi debug')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
import csv
import json
import os
import pprint
import sys

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
from pan_chainguard.ccadb import (revoked, valid_from_to,
                                  derived_trust_bits, TrustBits, TrustBitsMap2)


DEFAULT_POLICY = {
    'sources': ['mozilla'],
    'operation': 'union',
    'trust_bits': [],
}

SOURCES_MAP = {
    'mozilla': 'Mozilla',
    'google': 'Google Chrome',
    'apple': 'Apple',
    'microsoft': 'Microsoft'
}

CCADB_MAP = {v: k for k, v in SOURCES_MAP.items()}

args = None


def main():
    global args
    args = parse_args()

    asyncio.run(main_loop())

    sys.exit(0)


async def main_loop():
    certs = read_certs()
    policy = get_policy()

    if args.debug > 1:
        print('total', len(certs), file=sys.stderr)
    if args.debug > 2:
        print(pprint.pformat(certs, width=160), file=sys.stderr)
    if args.verbose:
        print('policy:', policy)

    if args.stats:
        stats(certs, policy)
        return

    policy_certs = get_certs(certs, policy)

    if args.verbose:
        print("%s: %d total certificates" % (
            ', '.join(policy['sources']), len(policy_certs)))

    if args.fingerprints is not None:
        write_fingerprints(certs, policy_certs)


def read_certs():
    certs = {}

    try:
        with open(args.ccadb, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile,
                                    dialect='unix')
            for row in reader:
                if row['Certificate Record Type'] != 'Root Certificate':
                    continue

                sha256 = row['SHA-256 Fingerprint']
                name = row['Certificate Name']

                ret, err = revoked(row)
                if ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    continue

                ret, err = valid_from_to(row)
                if not ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    continue

                certs[sha256] = row

    except OSError as e:
        print('%s: %s' % (args.ccadb, e), file=sys.stderr)
        sys.exit(1)

    return certs


def get_policy():
    def isvalid(policy):
        if not isinstance(policy['sources'], list):
            x = '"sources" not list'
        elif not len(policy['sources']):
            x = '"sources" empty'
        elif not all(item in SOURCES_MAP for item in policy['sources']):
            x = 'Invalid "sources"'
        elif policy['operation'] not in ['union', 'intersection']:
            x = 'Invalid "operation"'
        elif not all(item in TrustBitsMap2
                     for item in policy['trust_bits']):
            x = 'Invalid "trust_bits"'
        else:
            return True, None

        return False, x

    if args.policy is None:
        return DEFAULT_POLICY

    if os.path.isfile(args.policy):
        try:
            with open(args.policy, 'r') as f:
                x = json.load(f)
        except (OSError, ValueError) as e:
            print('%s: %s' % (args.policy, e), file=sys.stderr)
            sys.exit(1)

    else:
        try:
            x = json.loads(args.policy)
        except ValueError as e:
            print('%s: %s' % (e, args.policy), file=sys.stderr)
            sys.exit(1)

    if 'sources' not in x:
        x['sources'] = DEFAULT_POLICY['sources']
    if 'operation' not in x:
        x['operation'] = DEFAULT_POLICY['operation']
    if 'trust_bits' not in x:
        x['trust_bits'] = DEFAULT_POLICY['trust_bits']

    ret, e = isvalid(x)

    if not ret:
        msg = '%s: ' % args.policy if os.path.isfile(args.policy) else ''
        msg += '%s: %s' % (e, x)
        print(msg, file=sys.stderr)
        sys.exit(1)

    return x


def get_certs(certs, policy):
    certs_ = []

    for row in certs.values():
        sha256 = row['SHA-256 Fingerprint']

        if policy_match(policy, row):
            certs_.append(sha256)

    return certs_


def stats(certs, policy):
    sources = {}
    for x in SOURCES_MAP.keys():
        policy_ = {
            'sources': [x],
            'trust_bits': policy['trust_bits'],
            'operation': policy['operation'],
        }
        source_certs = get_certs(certs, policy_)

        sources[x] = set(source_certs)

    sets = []
    for x in sources.keys():
        print("%s: %d total certificates" % (x, len(sources[x])))
        sets.append(set(sources[x]))

    x = ', '.join(SOURCES_MAP.keys())

    intersection = set.intersection(*sets)
    print('%s: %d total certificates in all (intersection)' % (
        x, len(intersection)))

    union = set.union(*sets)
    print('%s: %d total certificates in any (union)' % (
        x, len(union)))


def policy_match(policy, row):
    def sources_match(source_cert, source_pol):
        if not source_cert:
            return False

        if policy['operation'] == 'union':
            for x in source_pol:
                if x in source_cert:
                    return True
            return False
        elif policy['operation'] == 'intersection':
            for x in source_pol:
                if x not in source_cert:
                    return False
            return True
        else:
            assert False, 'Invalid operation: %s' % policy['operation']

    def trust_bits_match(bits_cert, bits_pol):
        if bits_cert == TrustBits.NONE:
            return True

        for x in bits_pol:
            if TrustBitsMap2[x] not in bits_cert:
                return False
        return True

    certificate_name = row['Certificate Name']
    sha256 = row['SHA-256 Fingerprint']

    status_root = row['Status of Root Cert']
    ccadb_status = status_root.split(';')
    included = [x for x in ccadb_status if ': Included' in x]

    ccadb_sources = [x.split(':')[0].strip() for x in included]
    sources = [CCADB_MAP[x] for x in ccadb_sources if x in CCADB_MAP]

    trust_bits = derived_trust_bits(row)

    if args.debug > 1:
        print(certificate_name,
              'sources', sources,
              'trust-bits', trust_bits, file=sys.stderr)

    if (sources_match(sources, policy['sources']) and
       trust_bits_match(trust_bits, policy['trust_bits'])):
        return True

    if args.debug > 1:
        print('no match', certificate_name, sha256, file=sys.stderr)

    return False


def write_fingerprints(certs, policy_certs):
    # duplicate PAN-OS
    def canonicalize(x):
        x = ''.join(c for c in x if ord(c) < 128)
        x = x.replace(' ', '_')
        x = x.replace(',', '_')

        return x

    fieldnames = [
        'filename',
        'sha256',
    ]

    try:
        with open(args.fingerprints, 'w', newline='') as csvfile:
            writer = csv.DictWriter(csvfile,
                                    dialect='unix',
                                    fieldnames=fieldnames)
            writer.writeheader()
            sequence = 9000
            for k in policy_certs:
                row = {}
                row['sha256'] = k
                sequence += 1
                name = certs[k]['Certificate Name']
                row['filename'] = '%04d_%s' % (sequence,
                                               canonicalize(name))
                writer.writerow(row)

    except OSError as e:
        print('%s: %s' % (args.fingerprints, e), file=sys.stderr)
        sys.exit(1)


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='create custom root store from CCADB')
    # curl -OJ \
    # https://ccadb.my.salesforce-sites.com/ccadb/AllCertificateRecordsCSVFormatv2
    parser.add_argument('-c', '--ccadb',
                        required=True,
                        metavar='PATH',
                        help='CCADB AllCertificateRecordsCSVFormatv2 CSV path')
    parser.add_argument('-f', '--fingerprints',
                        metavar='PATH',
                        help='root store fingerprints CSV path')
    parser.add_argument('--policy',
                        metavar='JSON',
                        help='JSON policy object path or string')
    parser.add_argument('--stats',
                        action='store_true',
                        help='print source stats')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
import logging
import os
import re
import sys
import tarfile
import xml.etree.ElementTree as etree

try:
    import pan.xapi
except ImportError:
    print('Install pan-python: https://pypi.org/project/pan-python/',
          file=sys.stderr)
    sys.exit(1)

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import (title, __version__)

args = None


class Xpath():
    def __init__(self, *,
                 panorama=False,
                 template=None,
                 vsys=None):
        self.type = 'panorama' if panorama else 'firewall'
        if self.type == 'firewall' and template is not None:
            raise ValueError('template set for firewall')
        self.template = template
        self.vsys = vsys

    def __str__(self):
        x = self.__dict__
        return ', '.join((': '.join((k, str(x[k]))))
                         for k in sorted(x))

    @property
    def panorama(self):
        return self.type == 'panorama'

    def _variant(self, xpath):
        if self.type == 'firewall':
            if self.vsys is None:
                x = xpath[self.type]['shared']
            else:
                x = xpath[self.type]['vsys'] % self.vsys
        elif self.type == 'panorama':
            if self.template is None:
                x = xpath[self.type]['panorama']
            elif self.vsys is None:
                x = xpath[self.type]['template']['shared'] % self.template
            else:
                x = xpath[self.type]['template']['vsys'] % (
                    self.template, self.vsys)
        else:
            assert False, 'Malformed type: %s' % self.type

        return x

    def trusted_root_ca(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/ssl-decrypt/trusted-root-CA',
                'vsys': ("/config/devices/entry"
                         "[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/ssl-decrypt"
                         "/trusted-root-CA")
            },

            'panorama': {
                'panorama': '/config/panorama/ssl-decrypt/trusted-root-CA',
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']/config/shared"
                               "/ssl-decrypt/trusted-root-CA"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']/vsys"
                             "/entry[@name='%s']/ssl-decrypt/trusted-root-CA")
                },
            },
        }

        return self._variant(xpath)

    def certificates(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/certificate',
                'vsys': ("/config/devices/entry[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/certificate"),
            },

            'panorama': {
                'panorama': '/config/panorama/certificate',
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']"
                               "/config/shared/certificate"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']"
                             "/vsys/entry[@name='%s']/certificate")
                },
            },
        }

        return self._variant(xpath)

    def root_ca_exclude_list(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/ssl-decrypt/root-ca-exclude-list',
                'vsys': ("/config/devices/entry[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/ssl-decrypt/"
                         "root-ca-exclude-list"),
            },

            'panorama': {
                'panorama': ('/config/panorama/ssl-decrypt/'
                             'root-ca-exclude-list'),
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']"
                               "/config/shared/ssl-decrypt/"
                               "root-ca-exclude-list"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']"
                             "/vsys/entry[@name='%s']/ssl-decrypt/"
                             "root-ca-exclude-list")
                },
            },
        }

        return self._variant(xpath)


def main():
    global args
    args = parse_args()

    logger = logging.getLogger(pan.xapi.__name__)
    if args.xdebug == 3:
        logger.setLevel(pan.xapi.DEBUG3)
    elif args.xdebug == 2:
        logger.setLevel(pan.xapi.DEBUG2)
    elif args.xdebug == 1:
        logger.setLevel(pan.xapi.DEBUG1)

    log_format = '%(name)s: %(message)s'
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    asyncio.run(main_loop())

    sys.exit(0)


async def main_loop():
    xapi = None
    panorama = False

    if args.tag:
        try:
            xapi = pan.xapi.PanXapi(tag=args.tag,
                                    debug=args.xdebug)
            xapi.ad_hoc(modify_qs=True,
                        qs={'type': 'version'})
            elem = xapi.element_root.find('./result/model')
            if elem is not None:
                if elem.text == 'Panorama':
                    panorama = True
            else:
                print("Can't get model", file=sys.stderr)
        except pan.xapi.PanXapiError as e:
            print('pan.xapi.PanXapi:', e, file=sys.stderr)
            sys.exit(1)

    xpath = Xpath(panorama=panorama,
                  vsys=args.vsys,
                  template=args.template)
    if args.debug > 1:
        print('Xpath():', str(xpath), file=sys.stderr)
        print(xpath.certificates(), file=sys.stderr)
        print(xpath.trusted_root_ca(), file=sys.stderr)
        print(xpath.root_ca_exclude_list(), file=sys.stderr)

    if (any([args.disable_trusted, args.enable_trusted,
             args.delete, args.add, args.add_roots, args.commit]) and
       xapi is None):
        print('--tag argument required', file=sys.stderr)
        sys.exit(1)

    if args.enable_trusted:
        enable_trusted(xapi, xpath)

    if args.disable_trusted:
        disable_trusted(xapi, xpath)

    if args.delete:
        delete_certs(xapi, xpath)

    if args.add or args.add_roots:
        total = {
            'root': 0,
            'intermediate': 0,
        }
        try:
            with tarfile.open(name=args.certs, mode='r') as tar:
                for member in tar:
                    if args.verbose:
                        print('extracted %s' % member.name,
                              file=sys.stderr)
                    cert_name, cert_type = parse_cert(member.name)
                    f = tar.extractfile(member)
                    content = f.read()
                    if add_cert(xapi, xpath, cert_name, cert_type, content):
                        total[cert_type] += 1
        except (tarfile.TarError, OSError) as e:
            print("tarfile %s: %s" % (args.certs, e), file=sys.stderr)
            sys.exit(1)

        add_trusted_root_cas(xapi, xpath, add_cert.cert_names)

        for x in total:
            if total[x] > 0:
                print('%d %s certificates added' % (total[x], x))

    if args.commit:
        commit(xapi, panorama)


def enable_trusted(xapi, xpath):
    kwargs = {'xpath': xpath.root_ca_exclude_list()}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])
    entries = xapi.element_root.findall('./result/root-ca-exclude-list/member')
    if len(entries) > 0:
        api_request(xapi, xapi.delete, kwargs, 'success', ['7', '20'])
    print('%d default trusted root CAs enabled' % len(entries))


def disable_trusted(xapi, xpath):
    kwargs = {'xpath': '/config/predefined/trusted-root-ca'}
    api_request(xapi, xapi.get, kwargs, 'success', '19')

    total = 0
    entries = xapi.element_root.findall('./result/trusted-root-ca/entry')

    members = []
    for entry in entries:
        name = entry.attrib['name']
        member = '<member>%s</member>' % name
        members.append(member)

        total += 1
        if args.verbose:
            print('disabling', name, file=sys.stderr)

    root_ca_exclude = xpath.root_ca_exclude_list()
    element = ''.join(members)

    kwargs = {
        'xpath': root_ca_exclude,
        'element': element,
    }
    api_request(xapi, xapi.set, kwargs, 'success', '20')

    print('%d default trusted root CAs disabled' % total)


def parse_cert(path):
    pat = (r'^\d{4,4}_[().\w-]+/(intermediate|root)/'
           r'[0-9A-F]{64,64}\.crt$')
    if not re.search(pat, path):
        print('malformed path in archive: %s' % path,
              file=sys.stderr)
        sys.exit(1)

    head, crt = os.path.split(path)
    name, cert_type = os.path.split(head)
    cert_name = name[:4] + '-' + crt[:64]

    # PAN-OS certificate-name max len 63
    # Panorama certificate-name max len 31
    # PAN-99186 won't do
    return cert_name[:31], cert_type


def delete_certs(xapi, xpath):
    pat = r'^\d{4,4}-[0-9A-F]{26,26}$'

    certificates = xpath.certificates()
    kwargs = {'xpath': certificates}
    api_request(xapi, xapi.get, kwargs, 'success', '19')

    total = 0
    rootca = xpath.trusted_root_ca()
    entries = xapi.element_root.findall('./result/certificate/entry')

    for entry in entries:
        name = entry.attrib['name']
        if re.search(pat, name):
            member = "/member[text()='%s']" % name
            kwargs = {'xpath': rootca + member}
            api_request(xapi, xapi.delete, kwargs, 'success', ['7', '20'])

            entry = "/entry[@name='%s']" % name
            kwargs = {'xpath': certificates + entry}
            api_request(xapi, xapi.delete, kwargs, 'success', '20')

            total += 1
            if args.verbose:
                print('deleted', name, file=sys.stderr)

    print('%d certificates deleted' % total)


def add_cert(xapi, xpath, cert_name, cert_type, content):
    if (cert_type == 'root' and not args.add_roots or
       cert_type == 'intermediate' and not args.add):
        return False

    kwargs = {
        'category': 'certificate',
        'file': content,
        'extra_qs': {
            'certificate-name': cert_name,
            'format': 'pem',
        },
    }

    if xpath.panorama:
        if args.template is not None:
            kwargs['extra_qs']['target-tpl'] = args.template
        if args.vsys is not None:
            # XXX does not work
            kwargs['extra_qs']['target-tpl-vsys'] = args.vsys
    elif args.vsys is not None:
        kwargs['vsys'] = args.vsys

    try:
        xapi.import_file(**kwargs)
    except pan.xapi.PanXapiError as e:
        if 'Certificate is expired' in str(e):
            if args.verbose:
                print('%s expired' % cert_name, file=sys.stderr)
            return False
        else:
            print('%s: %s: %s' % (xapi.import_file.__name__, kwargs, e),
                  file=sys.stderr)
            sys.exit(1)

    # Use function attribute to cache certificate names so we can use
    # a single API request to enable them as trusted root CAs.
    try:
        add_cert.cert_names.append(cert_name)
    except AttributeError:
        add_cert.cert_names = [cert_name]

    if args.verbose:
        print('added %s %s' % (cert_type, cert_name), file=sys.stderr)

    return True


def add_trusted_root_cas(xapi, xpath, cert_names):
    members = []

    for x in cert_names:
        members.append('<member>%s</member>' % x)

    trustedca = xpath.trusted_root_ca()
    element = ''.join(members)

    kwargs = {
        'xpath': trustedca,
        'element': element,
    }

    api_request(xapi, xapi.set, kwargs, 'success', '20')


def api_request(xapi, func, kwargs, status=None, status_code=None):
    try:
        func(**kwargs)
    except pan.xapi.PanXapiError as e:
        print('%s: %s: %s' % (func.__name__, kwargs, e),
              file=sys.stderr)
        sys.exit(1)

    if status is not None and not s1_in_s2(xapi.status, status):
        print('%s: %s: status %s != %s' %
              (func.__name__, kwargs,
               xapi.status, status),
              file=sys.stderr)
        sys.exit(1)

    if (status_code is not None and
       not s1_in_s2(xapi.status_code, status_code)):
        print('%s: %s: status_code %s != %s' %
              (func.__name__, kwargs,
               xapi.status_code, status_code),
              file=sys.stderr)
        sys.exit(1)


def commit(xapi, panorama):
    root = etree.Element('commit')
    partial = etree.SubElement(root, 'partial')
    desc = etree.SubElement(partial, 'description')
    desc.text = '%s %s' % (title, __version__)
    if args.admin:
        admin = etree.SubElement(partial, 'admin')
        admin_member = etree.SubElement(admin, 'member')
        admin_member.text = args.admin

    policy_and_objects = etree.Element('policy-and-objects')
    policy_and_objects.text = 'excluded'
    device_and_network = etree.Element('device-and-network')
    device_and_network.text = 'excluded'
    shared_object = etree.Element('shared-object')
    shared_object.text = 'excluded'

    if panorama:
        if args.template:
            # commit scope: template
            # no template vsys scope
            template = etree.SubElement(partial, 'template')
            template_member = etree.SubElement(template, 'member')
            template_member.text = args.template
        else:
            # commit scope: device-and-network
            partial.append(shared_object)
    else:
        # firewall
        if args.vsys:
            # commit scope: vsys
            vsys = etree.SubElement(partial, 'vsys')
            vsys_member = etree.SubElement(vsys, 'member')
            vsys_member.text = args.vsys
            partial.append(device_and_network)
            partial.append(shared_object)
        else:
            # commit scope: shared-object
            partial.append(device_and_network)
            partial.append(policy_and_objects)

    cmd = etree.tostring(root).decode()
    if args.debug:
        print(cmd, file=sys.stderr)

    kwargs = {
        'cmd': cmd,
        'sync': True,
    }

    if args.verbose:
        print('commit config for admin %s' % args.admin)
    api_request(xapi, xapi.commit, kwargs, 'success')
    if args.debug:
        print(xapi.xml_root(), file=sys.stderr)

    if xapi.status_code is not None:
        code = ' [code=\"%s\"]' % xapi.status_code
    else:
        code = ''
    if xapi.status is not None:
        print('commit: %s%s' % (xapi.status, code), end='')
    if args.verbose and xapi.status_detail is not None:
        print(': "%s"' % xapi.status_detail.rstrip(), end='')
    print()


def s1_in_s2(s1, s2):
    if isinstance(s2, str):
        return s1 == s2
    elif isinstance(s2, list):
        return s1 in s2
    else:
        raise ValueError('Invalid type for s2. '
                         'Must be a string or a list of strings.')


def parse_args():
    def check_vsys(x):
        if x.isdigit():
            x = 'vsys' + x
        return x

    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='preload intermediate CAs')
    parser.add_argument('--tag', '-t',
                        help='.panrc tagname')
    parser.add_argument('--vsys',
                        type=check_vsys,
                        help='vsys name or number')
    parser.add_argument('--template',
                        help='Panorama template')
    x = 'certificates.tgz'
    parser.add_argument('--certs',
                        default=x,
                        metavar='PATH',
                        help='certificate archive path'
                        ' (default: %s)' % x)
    parser.add_argument('--add',
                        action='store_true',
                        help='add intermediate certificates')
    parser.add_argument('--add-roots',
                        action='store_true',
                        help='add root certificates')
    parser.add_argument('--delete',
                        action='store_true',
                        help='delete previously added certificates')
    parser.add_argument('--disable-trusted',
                        action='store_true',
                        help='disable all default trusted root CAs')
    parser.add_argument('--enable-trusted',
                        action='store_true',
                        help='enable all default trusted root CAs')
    parser.add_argument('--commit',
                        action='store_true',
                        help='commit configuration')
    parser.add_argument('--admin',
                        help='commit admin')
    parser.add_argument('--xdebug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='pan.xapi debug')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import aiohttp
import argparse
import asyncio
from collections import defaultdict
import csv
import io
import os
import pprint
import sys
import tarfile
import time

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
from pan_chainguard.ccadb import *
from pan_chainguard.crtsh import ArgsError, CrtShApi

ROOT = 'root'
INTERMEDIATE = 'intermediate'
MAX_TASKS = 3  # concurrent crt.sh API requests
CRT_SH_TIMEOUT = 60

args = None


def main():
    global args
    args = parse_args()

    ret = asyncio.run(main_loop())

    sys.exit(ret)


async def main_loop():
    create_archive(test=True)

    certs, invalid, warning, roots, intermediates, parents = get_certs()
    if args.debug:
        print('roots %d, intermediates %d, parents %d' %
              (len(roots), len(intermediates), len(parents)))

    chains = get_cert_chains(roots, intermediates, parents)
    if args.debug > 1:
        print('chains %d (root=>intermediates)' % len(chains),
              pprint.pformat(chains),
              file=sys.stderr)

    total, total_invalid, panos = await get_panos_intermediates(
        certs, chains, invalid, warning)
    print('%d invalid certificates found' % total_invalid)
    print('%d intermediate chains found for %d root CAs' % (
        len(panos), total))
    if args.debug > 1:
        print('intermediate chains', pprint.pformat(panos),
              file=sys.stderr)

    errors, certificates = await get_cert_files(panos)
    if errors:
        print('Error: %d certificates were not downloaded' % errors)
    else:
        print('All %d certificate chains were downloaded successfully'
              % len(certificates))

    if args.debug > 2:
        print('certificates', pprint.pformat(certificates),
              file=sys.stderr)

    create_archive(certificates)

    return 2 if errors else 0


def get_certs():
    invalid = {}
    warning = {}
    roots = []
    intermediates = []
    parents = defaultdict(list)
    certs = {}
    duplicates = defaultdict(list)

    try:
        with open(args.ccadb, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile,
                                    dialect='unix')
            for row in reader:
                sha256 = row['SHA-256 Fingerprint']
                name = row['Certificate Name']

                ret, err = revoked(row)
                if ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    invalid[sha256] = x
                    continue

                ret, err = valid_from_to(row)
                if not ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    invalid[sha256] = x
                    continue

                trust_bits = derived_trust_bits(row)
                if (trust_bits != TrustBits.NONE and
                   TrustBits.SERVER_AUTHENTICATION not in trust_bits):
                    x = 'Missing %s in %s %s' % (
                        TrustBits.SERVER_AUTHENTICATION.name,
                        trust_bits, sha256)
                    warning[sha256] = x

                cert_type = row['Certificate Record Type']

                if sha256 in certs:
                    if sha256 not in duplicates:
                        duplicates[sha256].append(certs[sha256])
                    duplicates[sha256].append(row)
                    continue

                certs[sha256] = row

                if cert_type == 'Root Certificate':
                    roots.append(sha256)

                if cert_type == 'Intermediate Certificate':
                    parent_sha256 = row['Parent SHA-256 Fingerprint']
                    if not parent_sha256:
                        x = 'Intermediate with no parent: %s' % sha256
                        if args.debug > 1:
                            print(x, file=sys.stderr)
                        invalid[sha256] = x
                        continue

                    if (trust_bits != TrustBits.NONE and
                       TrustBits.SERVER_AUTHENTICATION not in trust_bits):
                        x = 'Intermediate with no %s: %s' % \
                            (TrustBits.SERVER_AUTHENTICATION.name, sha256)
                        if args.debug > 1:
                            print(x, file=sys.stderr)
                        continue

                    intermediates.append(sha256)
                    parents[parent_sha256].append(sha256)

    except OSError as e:
        print('%s: %s' % (args.ccadb, e), file=sys.stderr)
        sys.exit(1)

    if args.debug:
        for sha256 in duplicates:
            print('Duplicate certificates %s' % sha256, file=sys.stderr)
            for x in duplicates[sha256]:
                print('  %s %s %s' % (x['Certificate Name'],
                                      x['Certificate Record Type'],
                                      x['Salesforce Record ID']),
                      file=sys.stderr)

    return certs, invalid, warning, roots, intermediates, parents


def get_cert_chains(roots, intermediates, parents):
    chains = defaultdict(list)

    for root in roots:
        if root not in parents:
            if args.debug:
                print('Root with no child %s' % root,
                      file=sys.stderr)
            continue

        for child in parents[root]:
            chain = []
            follow(chain, child, parents)
            chainlen = len(chain)
            if args.debug > 1:
                print('chain[%d]:' % chainlen,
                      pprint.pformat(chain), file=sys.stderr)

            chains[root].append(chain)

    return chains


def follow(chain, k, parents):
    if args.debug > 1:
        print('follow[%d]:' % len(chain), k, file=sys.stderr)

    chain.append(k)
    if k in parents:
        for child in parents[k]:
            follow(chain, child, parents)


async def get_panos_intermediates(certs, chains, invalid, warning):
    intermediates = {}
    not_in_common_store = {}
    total = 0
    total_invalid = 0

    if args.fingerprints:
        try:
            with open(args.fingerprints, 'r', newline='') as csvfile:
                reader = csv.DictReader(csvfile,
                                        dialect='unix')
                for row in reader:
                    sha256 = row['sha256']

                    if sha256 in warning:
                        print('Certificate warning %s: %s' % (
                            row['filename'],
                            warning[sha256]), file=sys.stderr)

                    if sha256 in invalid:
                        print('Invalid certificate %s: %s' % (
                            row['filename'],
                            invalid[sha256]), file=sys.stderr)
                        total_invalid += 1
                        continue

                    if sha256 not in certs:
                        print('Invalid certificate %s: %s' % (
                            row['filename'],
                            'Not found in CCADB'), file=sys.stderr)
                        total_invalid += 1
                        continue

                    status_root = certs[sha256]['Status of Root Cert']
                    statuses = status_root.split(';')
                    included = [': Included' in x for x in statuses]
                    if not any(included):
                        not_in_common_store[sha256] = status_root
                        print('Certificate not in common root store'
                              ' %s: %s' % (row['filename'], status_root),
                              file=sys.stderr)

                    total += 1
                    if sha256 in chains:
                        intermediates[row['filename']] = [(ROOT, sha256)]
                        for chain in chains[sha256]:
                            for x in chain:
                                intermediates[row['filename']].append(
                                    (INTERMEDIATE, x))
                    elif args.verbose:
                        x = '%s %s' % (row['filename'], 'intermediates 0')
                        if sha256 in not_in_common_store:
                            x += ' (not in common root store)'
                        print(x, file=sys.stderr)

        except OSError as e:
            print('%s: %s' % (args.fingerprints, e), file=sys.stderr)
            sys.exit(1)

    return total, total_invalid, intermediates


async def download(api, sequence, sha256):
    tries = 0
    MAX_TRIES = 5
    RETRY_SLEEP = 5.0

    RETRY_STATUS = [
        429,  # Too Many Requests
        502,  # Bad Gateway
        503,  # Service Unavailable
        504,  # Gateway Time-out
    ]

    while True:
        tries += 1
        try:
            resp = await api.download(id=sha256[1])

            if resp.status in RETRY_STATUS:
                x = '%d %s %s %s' % (
                    resp.status, resp.reason, sequence, sha256[1])
                if tries == MAX_TRIES:
                    print('no retry after try %d %s' % (tries, x),
                          file=sys.stderr)
                    x = 'download failed ' + x
                    break
                print('retry after try %d %s' % (tries, x),
                      file=sys.stderr)
                await asyncio.sleep(RETRY_SLEEP)
            elif resp.status != 200:
                x = 'download failed %d %s %s %s' % (
                    resp.status, resp.reason, sequence, sha256[1])
                break
            else:
                filename, content = await api.content(resp=resp)
                if filename is None:
                    filename = sha256[1] + '.crt'

                start = '-----BEGIN CERTIFICATE-----'
                end = '-----END CERTIFICATE-----\n'
                if (content.startswith(start) and
                   content.endswith(end)):
                    return filename, content, sequence, sha256[0]
                else:
                    # XXX ephemeral?
                    x = 'content malformed %s %s %s' % (
                        sequence, sha256[1], content)
                    if tries == MAX_TRIES:
                        print('no retry after try %d %s' % (tries, x),
                              file=sys.stderr)
                        x = 'download failed ' + x
                        break
                    print('retry after try %d %s' % (tries, x),
                          file=sys.stderr)
                    await asyncio.sleep(RETRY_SLEEP)

        except (asyncio.TimeoutError,
                asyncio.CancelledError,  # XXX
                aiohttp.ClientError) as e:
            msg = e if str(e) else type(e).__name__
            x = 'CrtShApi: %s %s %s' % (msg, sequence, sha256[1])
            if tries == MAX_TRIES:
                print('no retry after try %d %s' % (tries, x),
                      file=sys.stderr)
                x = 'download failed ' + x
                break
            print('retry after try %d %s' % (tries, x),
                  file=sys.stderr)
            # no sleep

        except ArgsError as e:
            x = 'CrtShApi: %s' % e
            break

    return None, x, None, None


async def get_cert_files(panos):
    user_agent = '%s/%s' % (title, __version__)
    headers = {'user-agent': user_agent}

    try:
        api = CrtShApi(timeout=CRT_SH_TIMEOUT, headers=headers)
    except ArgsError as e:
        print('CrtShApi: %s' % e, file=sys.stderr)
        sys.exit(1)
    else:
        errors, certificates = await process_roots(api, panos)
    finally:
        await api.session.close()

    return errors, certificates


async def process_roots(api, roots):
    certificates = defaultdict(list)
    tasks = []
    start = time.time()
    total = 0
    errors = 0

    for sequence in roots:
        sequence_friendly = sequence
        if sequence.endswith('.cer'):
            sequence_friendly = sequence[:-4]
        if args.verbose:
            print('download %s intermediates %d' % (
                sequence_friendly,
                len(roots[sequence][1:])))

        for sha256 in roots[sequence]:
            if not args.roots and sha256[0] == ROOT:
                continue
            total += 1
            tasks.append(download(api, sequence_friendly, sha256))

            if len(tasks) == MAX_TASKS:
                errors += await run_tasks(tasks, certificates)
                tasks = []

    if len(tasks):
        errors += await run_tasks(tasks, certificates)

    if args.debug:
        end = time.time()
        elapsed = end - start
        mins = elapsed // 60
        secs = elapsed % 60
        avg = elapsed / total if total > 0 else 0
        print('%d tasks, elapsed %.2f seconds, %dm%ds, avg %.2fs' %
              (total, elapsed, mins, secs, avg),
              file=sys.stderr)

    return errors, certificates


async def run_tasks(tasks, certificates):
    INTERVAL_WAIT = 3.3  # random throttle sweet spot to avoid TimeoutError
    errors = 0

    if args.debug:
        print('running %d tasks' % len(tasks),
              file=sys.stderr)
        start = time.time()

    for coro in asyncio.as_completed(tasks):
        filename, content, sequence, cert_type = await coro
        if filename is None:
            # error
            print(content, file=sys.stderr)
            errors += 1
        else:
            certificates[sequence].append((cert_type, filename, content))

    if args.debug:
        end = time.time()
        elapsed = end - start
        rate = len(tasks) / elapsed
        rate2 = elapsed / len(tasks)
        print('%d tasks complete, elapsed %.2f seconds, '
              '%.2f tasks/sec, %.2f secs/task' %
              (len(tasks), elapsed, rate, rate2), file=sys.stderr)
        print('sleeping %.2fs' % INTERVAL_WAIT, file=sys.stderr)
        await asyncio.sleep(INTERVAL_WAIT)

    return errors


def create_archive(certificates=None, test=False):
    if not certificates:
        return

    try:
        with tarfile.open(name=args.certs, mode='w:gz') as tar:
            if test:
                os.unlink(args.certs)
                return
            for sequence in sorted(certificates.keys()):
                for cert in certificates[sequence]:
                    name = os.path.join(sequence, cert[0], cert[1])
                    member = tarfile.TarInfo(name=name)
                    member.size = len(cert[2])
                    member.mtime = time.time()
                    f = io.BytesIO(cert[2].encode())
                    tar.addfile(member, fileobj=f)
    except (tarfile.TarError, OSError) as e:
        print("tarfile %s: %s" % (args.certs, e), file=sys.stderr)
        sys.exit(1)


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='generate intermediate CAs to preload')
    # curl -OJ \
    # https://ccadb.my.salesforce-sites.com/ccadb/AllCertificateRecordsCSVFormatv2
    parser.add_argument('-c', '--ccadb',
                        required=True,
                        metavar='PATH',
                        help='CCADB AllCertificateRecordsCSVFormatv2 CSV path')
    # sh cert-fingerprints.sh cert-dir
    parser.add_argument('-f', '--fingerprints',
                        metavar='PATH',
                        help='root CAs fingerprints CSV path')
    x = 'certificates.tgz'
    parser.add_argument('--certs',
                        default=x,
                        metavar='PATH',
                        help='certificate archive path'
                        ' (default: %s)' % x)
    parser.add_argument('--roots',
                        action='store_true',
                        help='also download root CAs')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()
